]> git.sur5r.net Git - u-boot/blobdiff - cmd/elf.c
bootvx: x86: Make VxWorks EFI console driver happy
[u-boot] / cmd / elf.c
index 5d071ee26c44b65b293ca0897c173644900cd3d1..30c845833b6f6030545bb33a273f061867e0b3c1 100644 (file)
--- a/cmd/elf.c
+++ b/cmd/elf.c
 #include <common.h>
 #include <command.h>
 #include <elf.h>
+#include <environment.h>
 #include <net.h>
 #include <vxworks.h>
 #ifdef CONFIG_X86
+#include <vbe.h>
 #include <asm/e820.h>
 #include <linux/linkage.h>
 #endif
 
 /*
* A very simple elf loader, assumes the image is valid, returns the
  * entry point address.
+ *
+ * Note if U-Boot is 32-bit, the loader assumes the to segment's
+ * physical address and size is within the lower 32-bit address space.
+ */
+static unsigned long load_elf64_image_phdr(unsigned long addr)
+{
+       Elf64_Ehdr *ehdr; /* Elf header structure pointer */
+       Elf64_Phdr *phdr; /* Program header structure pointer */
+       int i;
+
+       ehdr = (Elf64_Ehdr *)addr;
+       phdr = (Elf64_Phdr *)(addr + (ulong)ehdr->e_phoff);
+
+       /* Load each program header */
+       for (i = 0; i < ehdr->e_phnum; ++i) {
+               void *dst = (void *)(ulong)phdr->p_paddr;
+               void *src = (void *)addr + phdr->p_offset;
+
+               debug("Loading phdr %i to 0x%p (%lu bytes)\n",
+                     i, dst, (ulong)phdr->p_filesz);
+               if (phdr->p_filesz)
+                       memcpy(dst, src, phdr->p_filesz);
+               if (phdr->p_filesz != phdr->p_memsz)
+                       memset(dst + phdr->p_filesz, 0x00,
+                              phdr->p_memsz - phdr->p_filesz);
+               flush_cache((unsigned long)dst, phdr->p_filesz);
+               ++phdr;
+       }
+
+       return ehdr->e_entry;
+}
+
+/*
+ * A very simple ELF loader, assumes the image is valid, returns the
+ * entry point address.
+ *
+ * The loader firstly reads the EFI class to see if it's a 64-bit image.
+ * If yes, call the ELF64 loader. Otherwise continue with the ELF32 loader.
  */
 static unsigned long load_elf_image_phdr(unsigned long addr)
 {
@@ -34,12 +74,16 @@ static unsigned long load_elf_image_phdr(unsigned long addr)
        int i;
 
        ehdr = (Elf32_Ehdr *)addr;
+       if (ehdr->e_ident[EI_CLASS] == ELFCLASS64)
+               return load_elf64_image_phdr(addr);
+
        phdr = (Elf32_Phdr *)(addr + ehdr->e_phoff);
 
        /* Load each program header */
        for (i = 0; i < ehdr->e_phnum; ++i) {
                void *dst = (void *)(uintptr_t)phdr->p_paddr;
                void *src = (void *)addr + phdr->p_offset;
+
                debug("Loading phdr %i to 0x%p (%i bytes)\n",
                      i, dst, phdr->p_filesz);
                if (phdr->p_filesz)
@@ -208,8 +252,11 @@ int do_bootvx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
        char build_buf[128]; /* Buffer for building the bootline */
        int ptr = 0;
 #ifdef CONFIG_X86
-       struct e820info *info;
-       struct e820entry *data;
+       ulong base;
+       struct e820_info *info;
+       struct e820_entry *data;
+       struct efi_gop_info *gop;
+       struct vesa_mode_info *vesa = &mode_info.vesa;
 #endif
 
        /*
@@ -240,13 +287,9 @@ int do_bootvx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
         * from the VxWorks BSP header files.
         * This will vary from board to board
         */
-#if defined(CONFIG_WALNUT)
-       tmp = (char *)CONFIG_SYS_NVRAM_BASE_ADDR + 0x500;
-       eth_getenv_enetaddr("ethaddr", (uchar *)build_buf);
-       memcpy(tmp, &build_buf[3], 3);
-#elif defined(CONFIG_SYS_VXWORKS_MAC_PTR)
+#if defined(CONFIG_SYS_VXWORKS_MAC_PTR)
        tmp = (char *)CONFIG_SYS_VXWORKS_MAC_PTR;
-       eth_getenv_enetaddr("ethaddr", (uchar *)build_buf);
+       eth_env_get_enetaddr("ethaddr", (uchar *)build_buf);
        memcpy(tmp, build_buf, 6);
 #else
        puts("## Ethernet MAC address not copied to NV RAM\n");
@@ -300,7 +343,7 @@ int do_bootvx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
                                ptr += sprintf(build_buf + ptr, "e=%s", tmp);
                                tmp = env_get("netmask");
                                if (tmp) {
-                                       u32 mask = getenv_ip("netmask").s_addr;
+                                       u32 mask = env_get_ip("netmask").s_addr;
                                        ptr += sprintf(build_buf + ptr,
                                                       ":%08x ", ntohl(mask));
                                } else {
@@ -338,25 +381,41 @@ int do_bootvx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
 
 #ifdef CONFIG_X86
        /*
-        * Since E820 information is critical to the kernel, if we don't
-        * specify these in the environments, use a default one.
+        * Get VxWorks's physical memory base address from environment,
+        * if we don't specify it in the environment, use a default one.
         */
-       tmp = env_get("e820data");
-       if (tmp)
-               data = (struct e820entry *)simple_strtoul(tmp, NULL, 16);
-       else
-               data = (struct e820entry *)VXWORKS_E820_DATA_ADDR;
-       tmp = env_get("e820info");
-       if (tmp)
-               info = (struct e820info *)simple_strtoul(tmp, NULL, 16);
-       else
-               info = (struct e820info *)VXWORKS_E820_INFO_ADDR;
+       base = env_get_hex("vx_phys_mem_base", VXWORKS_PHYS_MEM_BASE);
+       data = (struct e820_entry *)(base + E820_DATA_OFFSET);
+       info = (struct e820_info *)(base + E820_INFO_OFFSET);
 
-       memset(info, 0, sizeof(struct e820info));
+       memset(info, 0, sizeof(struct e820_info));
        info->sign = E820_SIGNATURE;
        info->entries = install_e820_map(E820MAX, data);
-       info->addr = (info->entries - 1) * sizeof(struct e820entry) +
-                    VXWORKS_E820_DATA_ADDR;
+       info->addr = (info->entries - 1) * sizeof(struct e820_entry) +
+                    E820_DATA_OFFSET;
+
+       /*
+        * Explicitly clear the bootloader image size otherwise if memory
+        * at this offset happens to contain some garbage data, the final
+        * available memory size for the kernel is insane.
+        */
+       *(u32 *)(base + BOOT_IMAGE_SIZE_OFFSET) = 0;
+
+       /*
+        * Prepare compatible framebuffer information block.
+        * The VESA mode has to be 32-bit RGBA.
+        */
+       if (vesa->x_resolution && vesa->y_resolution) {
+               gop = (struct efi_gop_info *)(base + EFI_GOP_INFO_OFFSET);
+               gop->magic = EFI_GOP_INFO_MAGIC;
+               gop->info.version = 0;
+               gop->info.width = vesa->x_resolution;
+               gop->info.height = vesa->y_resolution;
+               gop->info.pixel_format = EFI_GOT_RGBA8;
+               gop->info.pixels_per_scanline = vesa->bytes_per_scanline / 4;
+               gop->fb_base = vesa->phys_base_ptr;
+               gop->fb_size = vesa->bytes_per_scanline * vesa->y_resolution;
+       }
 #endif
 
        /*
@@ -365,13 +424,18 @@ int do_bootvx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
         * binary image.
         */
        if (valid_elf_image(addr))
-               addr = load_elf_image_shdr(addr);
+               addr = load_elf_image_phdr(addr);
        else
                puts("## Not an ELF image, assuming binary\n");
 
        printf("## Starting vxWorks at 0x%08lx ...\n", addr);
 
        dcache_disable();
+#if defined(CONFIG_ARM64) && defined(CONFIG_ARMV8_PSCI)
+       armv8_setup_psci();
+       smp_kick_all_cpus();
+#endif
+
 #ifdef CONFIG_X86
        /* VxWorks on x86 uses stack to pass parameters */
        ((asmlinkage void (*)(int))addr)(0);
@@
 
 /*
- * A very simple elf loader, assumes the image is valid, returns the
+ * A very simple ELF64 loader, assumes the image is valid, returns the
  * entry point address.